import matplotlib.text as mtext
import matplotlib.patches as mpatches
import matplotlib.transforms as mtransforms
from matplotlib.path import Path
from matplotlib.patches import bbox_artist
import math

def get_textbox(text, renderer):
    """figure out the bounding box of the text. Unlike get_extents()
       method, bbox of the text before the rotation is returned"""

    projected_xs = []
    projected_ys = []
        
    theta = text.get_rotation()/180.*math.pi
    tr = mtransforms.Affine2D().rotate(-theta)

    for t, wh, x, y in text._get_layout(renderer)[1]:
        w, h = wh


        xt1, yt1 = tr.transform_point((x, y))
        xt2, yt2 = xt1+w, yt1+h
            
        projected_xs.extend([xt1, xt2])
        projected_ys.extend([yt1, yt2])


    xt_box, yt_box = min(projected_xs), min(projected_ys)
    w_box, h_box = max(projected_xs) - xt_box, max(projected_ys) - yt_box

    tr = mtransforms.Affine2D().rotate(theta)

    x_box, y_box = tr.transform_point((xt_box, yt_box))

    return x_box, y_box, w_box, h_box


class NotImplemented(Exception):
    pass

class TextBoxBase:
    """ Base class
    """
    def __init__(self, **kw):
        self.kw = kw

    def __call__(self, fontsize, x0, y0, width, height):
        """ This method needs to be overriden in its subclass. It has
        to return a Path instance around the bbox(x0, y0, width,
        height). The fontsize argument is meant to be used for a
        fontsize-proportional padding.
        """
        raise NotImplemented("Subclass of TextBoxBase have to overrid __call__ method")
    
    def set_pad(self, pad):
        if pad is not None:
            self.pad = pad


    def get_kw(self):
        "This method is not used currently"
        return self.kw


class SquareBox(TextBoxBase):
    """ Simple square box.
         'pad' :an amount of padding around the text. '1' means fontsize. """
    
    def __init__(self, pad=0.2, **kw):
        self.pad = pad
        TextBoxBase.__init__(self, **kw)

    def __call__(self, fontsize, x0, y0, width, height):

        # padding
        _pad_box = self.pad * fontsize

        # width and height with padding added.
        width, height = width + 2.*_pad_box, \
                        height + 2.*_pad_box,

        # boundary of the padded box
        x0, y0 = x0-_pad_box, y0-_pad_box, 
        x1, y1 = x0+width, y0 + height

        cp = [(x0, y0),
              (x1, y0),
              (x1, y1),
              (x0, y1),
              (x0, y0),
              (x0, y0)]

        com = [Path.MOVETO,
               Path.LINETO,
               Path.LINETO,
               Path.LINETO,
               Path.LINETO,
               Path.CLOSEPOLY]

        text_box_path = Path(cp, com)
        
        return text_box_path


class RoundBox(TextBoxBase):
    """ A box with round corners.
    """
    
    def __init__(self, pad=0.2, **kw):
        TextBoxBase.__init__(self, **kw)
        self.set_pad(pad)

    def __call__(self, fontsize, x0, y0, width, height):
        
        _pad_box = self.pad * fontsize
        width, height = width + 2.*_pad_box, \
                        height + 2.*_pad_box,

        x0, y0 = x0-_pad_box, y0-_pad_box, 
        x1, y1 = x0+width, y0 + height

        # Round corners are implemented as quadratic bezier. eg.
        # [(x0, y0-dr), (x0, y0), (x0+dr, y0)] for lower left corner.
        dr = _pad_box
        cp = [(x0+dr, y0),
              (x1-dr, y0),
              (x1, y0), (x1, y0+dr),
              (x1, y1-dr),
              (x1, y1), (x1-dr, y1),
              (x0+dr, y1),
              (x0, y1), (x0, y1-dr),
              (x0, y0+dr),
              (x0, y0), (x0+dr, y0),
              (x0+dr, y0)]

        com = [Path.MOVETO,
               Path.LINETO,
               Path.CURVE3, Path.CURVE3,
               Path.LINETO,
               Path.CURVE3, Path.CURVE3,
               Path.LINETO,
               Path.CURVE3, Path.CURVE3,
               Path.LINETO,
               Path.CURVE3, Path.CURVE3,
               Path.CLOSEPOLY]
               
        text_box_path = Path(cp, com)
        
        return text_box_path




class SBox(TextBoxBase):
    """ """
    def __init__(self, pad=0.2, **kw):
        self.pad = pad
        TextBoxBase.__init__(self, **kw)

    def set_pad(self, pad):
        if pad is not None:
            self.pad = pad

    def __call__(self, fontsize, x0, y0, width, height):
        
        _pad_box = self.pad * fontsize
        width, height = width + 2.*_pad_box, \
                        height + 2.*_pad_box,

        x0, y0 = x0-_pad_box, y0-_pad_box, 
        x1, y1 = x0+width, y0 + height

        dx = .5*_pad_box
        dy = y1 - y0
        cp = [(x0, y0),
              (x1, y0),
              (x1+dx, y0+0.25*dy), (x1, y0+0.5*dy),
              (x1-dx, y0+0.75*dy), (x1, y1),
              (x0, y1),
              (x0-dx, y0+0.75*dy), (x0, y0+0.5*dy),
              (x0+dx, y0+0.25*dy), (x0, y0),
              (x0, y0)]

        com = [Path.MOVETO,
               Path.LINETO,
               Path.CURVE3, Path.CURVE3,
               Path.CURVE3, Path.CURVE3,
               Path.LINETO,
               Path.CURVE3, Path.CURVE3,
               Path.CURVE3, Path.CURVE3,
               Path.CLOSEPOLY]
               
        text_box_path = Path(cp, com)
        
        return text_box_path



class MoreRoundBox(TextBoxBase):
    def __init__(self, pad=0.2, rounding_fraction=1., **kw):
        self.pad = pad
        self.rounding_fraction = rounding_fraction
        TextBoxBase.__init__(self, **kw)

    def set_pad(self, pad):
        if pad is not None:
            self.pad = pad

    def __call__(self, fontsize, x0, y0, width, height):
        
        _pad_box = self.pad * fontsize
        width, height = width + 2.*_pad_box, \
                        height + 2.*_pad_box,

        x0, y0 = x0-_pad_box, y0-_pad_box, 
        x1, y1 = x0+width, y0 + height

        dr = _pad_box*self.rounding_fraction
        cp = [(x0, y0),
              (x0+dr, y0-dr), (x1-dr, y0-dr), (x1, y0),
              (x1+dr, y0+dr), (x1+dr, y1-dr), (x1, y1),
              (x1-dr, y1+dr), (x0+dr, y1+dr), (x0, y1),
              (x0-dr, y1-dr), (x0-dr, y0+dr), (x0, y0),
              (x0, y0)]

        com = [Path.MOVETO,
               Path.CURVE4, Path.CURVE4, Path.CURVE4, 
               Path.CURVE4, Path.CURVE4, Path.CURVE4, 
               Path.CURVE4, Path.CURVE4, Path.CURVE4, 
               Path.CURVE4, Path.CURVE4, Path.CURVE4, 
               Path.CLOSEPOLY]
               
        text_box_path = Path(cp, com)
        
        return text_box_path



class TextBoxArtist:
    """ A textbox artist. The calling convention for the __call__
    method are borrowed from bbox_artist function."""
    
    def __init__(self, path_maker):
        """ path_maker must be a isntance of TextBoxBase or similar,
        i.e., a callable object which returns a Path of the
        box and has a set_pad() method."""
        
        self.path_maker = path_maker
        
    def __call__(self, artist, renderer, props={}, fill=True):

        # as we use "get_textbox" function, only text instances are supproted.
        # We may fall back to get_extents() methods if not.
        assert isinstance(artist, mtext.Text)
        
        pad = props.pop("pad", None)
        self.path_maker.set_pad(pad)

        # properties for a shadow.
        shadow_props = props.pop("shadow", {})

        text = artist

        # x_box, y_box : lower-left corrdinate of the textbox. The textbox is
        # also rotated around this point.
        x_box, y_box, w_box, h_box = get_textbox(text, renderer)

        # postion of the text in the display coordinate.
        x, y = text.get_position()
        x, y = text.get_transform().transform_point((x, y))

        # transformation for the textbox.
        tr = mtransforms.Affine2D().translate(-x_box, -y_box)
        theta = text.get_rotation()/180.*math.pi
        tr = tr.rotate(theta).translate(x_box, y_box)
        tr.translate(x, y)

        # textbox patch
        path = self.path_maker(text.get_size(), x_box,y_box, w_box, h_box)
        patch = mpatches.PathPatch(path, 
                                   transform=tr, **props)

        # shadow
        if shadow_props:
            ox, oy = shadow_props.pop("offset", (5, -5))
            shadow = mpatches.Shadow(patch, ox, oy,
                                     transform=tr, props=shadow_props)
            shadow.draw(renderer)
            
        patch.draw(renderer)

        


class Text(mtext.Text):
    def __init__(self, *kl, **kw):
        self._textbox_props = kw.pop("bbox", None)
        mtext.Text.__init__(self, *kl, **kw)

        
    def draw(self, renderer):
        if self._textbox_props is not None:
            props = self._textbox_props.copy()
            textbox_artist = props.pop("artist", bbox_artist)
            textbox_artist(self, renderer, props)

        mtext.Text.draw(self, renderer)




